from dataclasses import dataclass
from typing import Literal, Union

import numpy as np


@dataclass
class Outcomes:
    peak_to_peak: Union[float, None] = None
    area: Union[float, None] = None
    onset: Union[float, None] = None


@dataclass
class Single:
    waveform: np.array = None
    outcomes: Union[dict, None] = None
    background_sd: float = None
    stim_intensity: int = None
    extract_indexes: tuple = None
    extract_times: tuple = None


@dataclass
class Singles:
    x_axis_extract: np.array = None
    reflexes: list[Single] = None
    avg_waveform: dict = None
    mean_outcomes: Union[list, None] = None
    type: Literal["single", "double", "train"] = None
    sd_window_idx: list = None
    reflex_windows_idx: dict = None
    sd_window_ms: list = None
    reflex_windows_ms: dict = None


@dataclass
class Double:
    waveform: np.array = None
    reflex1: Single = None
    reflex2: Single = None
    ratio: Union[dict, None] = None
    stim_intensity: int = None
    extract_indexes: tuple = None
    extract_times: tuple = None


@dataclass
class Doubles:
    x_axis_extract: np.array = None
    x_axis_singles: np.array = None
    reflexes: list[Double] = None
    avg_reflex1: Single = None
    avg_reflex1_for_doubles: Single = None
    avg_reflex2: Single = None
    avg_waveform = None
    avg_waveform_for_doubles = None
    mean_outcomes_reflex1: Union[list, None] = None
    mean_outcomes_reflex2: Union[list, None] = None
    mean_ratio: float = None
    type: Literal["single", "double", "train"] = None
    sd_window_idx: list = None
    reflex_windows_idx: dict = None
    sd_window_ms: list = None
    reflex_windows_ms: dict = None


class SectionReflexes:
    def __init__(self, info, reflexes):
        self.info = info
        self.reflexes: dict = reflexes



import spike2py_reflex as s2pr


def reflexes(section):
    """Generate raw data plots of all reflexes in a trial section

    Included are the individual traces and the mean trace, as well
    as windows indicating the region where reflex outcomes were
    calculated.
    """

    muscles = list(section.reflexes.keys())
    if section.reflexes[muscles[0]].type == s2pr.utils.SINGLE:
        s2pr.plot.plot_single(section)
        s2pr.plot.single_outcomes(section)
    elif section.reflexes[muscles[0]].type == s2pr.utils.DOUBLE:
        s2pr.plot.plot_double(section)
        s2pr.plot.double_outcomes(section)
    elif section.reflexes[muscles[0]].type == s2pr.utils.TRAIN:
        s2pr.plot.plot_train(section)
        s2pr.plot.train_outcomes(section)





import spike2py_reflex as s2pr


def test_reflex_extract_hreflex_doubles(info_data_hreflex):
    info, data = info_data_hreflex
    section = s2pr.reflexes.extract(info, data)
    assert section.info.section == 'hreflex'
    assert list(section.reflexes.keys())[0] == 'Fdi'
    assert len(section.reflexes['Fdi'].reflexes) == 7
    assert len(section.reflexes['Fdi'].reflexes[0].waveform) == 794
    assert len(section.reflexes['Fdi'].reflexes[0].reflex1.waveform) == 109
    assert len(section.reflexes['Fdi'].reflexes[0].reflex2.waveform) == 109
    assert len(section.reflexes['Fdi'].x_axis_extract) == 794
    assert len(section.reflexes['Fdi'].x_axis_singles) == 109


def test_reflex_extract_ramp_singles(info_data_ramp):
    info, data = info_data_ramp
    section = s2pr.reflexes.extract(info, data)
    assert section.info.section == 'ramp'
    assert list(section.reflexes.keys())[0] == 'Fdi'
    assert len(section.reflexes['Fdi'].reflexes) == 1661
    assert len(section.reflexes['Fdi'].reflexes[0].waveform) == 90
    assert len(section.reflexes['Fdi'].x_axis_extract) == 90


def test_reflex_extract_mmax_singles(info_data_mmax):
    info, data = info_data_mmax
    section = s2pr.reflexes.extract(info, data)
    assert section.info.section == 'mmax'
    assert list(section.reflexes.keys())[0] == 'Fdi'
    assert len(section.reflexes['Fdi'].reflexes) == 6
    assert len(section.reflexes['Fdi'].reflexes[0].waveform) == 694
    assert len(section.reflexes['Fdi'].x_axis_extract) == 694



from pathlib import Path
from typing import Union
import pickle

import spike2py as s2p
import spike2py_reflex as s2pr
import spike2py_preprocess as s2pp


def study(study_path: Path, plot=False):
    """Extract reflexes for all trials of all subject for a study

    Parameters
    ----------
    study_path: Path to study folder. Folder is to be structured as indicated in spike2py_preprocess docs
    plot: Flag indicating whether to generate plots
    """
    info = s2pr.info.Info(study_path, plot)
    for subject_ in info.subjects:
        subject(subject_, info=info, plot=plot)


def subject(
    subject_: str,
    info: Union[s2pr.info.Info, None] = None,
    study_path: Union[str, Path, None] = None,
    plot: bool = False,
    from_command_line: bool = False,
):
    """Process all trials and their sections for a given subject

    Parameters
    ----------
    subject_: Subject id, which also has to be the name of their folder
    study_path: Path to study folder; required if called via __main__ interface
    info: Study-related details; required if called internally
    plot: Whether to generate plots
    from_command_line: Whether function being called via __main__ interface

    """
    if from_command_line:
        study_path = Path(study_path)
        info = s2pr.info.Info(study_path, plot)
    info.clear_subject()
    info.init_subject(subject_)

    # Iterate over each trial for the subject
    for trial_, sections in info.trials_sections.trials_sections.items():
        trial(trial_=trial_, info=info, plot=plot, sections=sections)

    if plot:
        if study_path is None:
            study_path = info.study_path
        subject_path = study_path / subject_ / "figures" / "reflexes"
        s2pp.utils.merge_pdfs(subject_path)
        subject_path = study_path / subject_ / "figures" / "outcomes"
        s2pp.utils.merge_pdfs(subject_path)


def trial(
    trial_: str,
    subject_: Union[str, None] = None,
    info: Union[s2pr.info.Info, None] = None,
    sections: Union[list, None] = None,
    study_path: Union[str, Path, None] = None,
    plot: bool = False,
    from_command_line: bool = False,
):
    """Process a given trial and its various sections

    Parameters
    ----------
    trial_: name of trial, as indicated in the associated .pkl file
    subject_: subject id; required when processing indivdual trial
    info: info related to study/subject; if not provided because processing a single trial,
          then it will be created.
    section: sections that need to be extracted and processed in given trial
    study_path: absolute path to top-level study folder; required when processing a
                single trial
    plot: flag to determine whether to generate plots
    from_command_line: flag indentifying whether processing single trial (from command line
                       usually)
    """

    if from_command_line:
        study_path = Path(study_path)
        info = s2pr.info.Info(study_path, plot)
        info.clear_subject()
        info.init_subject(subject_)

    info.trial = trial_

    for section_ in sections:
        try:
            info.clear_section()
            info.init_section(section_)
            print(f"\tProcessing: {info.trial}-{section_}")
            print("\t\t", info.section_pkl)
            data = s2p.trial.load(info.section_pkl)
            info.triggers = s2pr.utils.Triggers(info, data)

            section = s2pr.reflexes.extract(info, data)
            section = s2pr.outcomes.calculate(section)
            s2pr.outcomes.summary(section)

            if info.plot:
                s2pr.plot.reflexes(section)
                s2pr.plot.outcomes(section)

            _save_section_reflexes(section)
        except FileNotFoundError:
            print("\t\t Does not exists! Make sure this is expected.")


def _save_section_reflexes(section):
    section_pkl = section.info.section_pkl
    file_name = section_pkl.name
    file_name_parts = file_name.split(".")
    section_reflex_pkl_file_name = (
        file_name_parts[0] + "_reflexes." + file_name_parts[1]
    )
    section_reflex_pkl = section_pkl.parent / section_reflex_pkl_file_name
    with open(section_reflex_pkl, "wb") as output:
        pickle.dump(section, output, pickle.HIGHEST_PROTOCOL)



from .constants import (SINGLE,
                        DOUBLE,
                        TRAIN,
                        TRAIN_SINGLE_PULSE,
                        CONVERT_MS_TO_S,
                        KHZ_ISI_THRESHOLD,
                        SD_MULTIPLIER,
                        STUDY_INFO_FILE,
                        REFLEX_FILE,
                        SUBJECT_REFLEX_FILE)
from .stim_intensity import get_stim_intensity
from .triggers import Triggers



import spike2py_reflex as s2pr


def test_init():
    ts = s2pr.info.TrialsSections()
    assert ts._study is None
    assert ts._subject is None
    assert ts._section is None


def test_get_attributes_simple(demo_study_reflex):
    study_ts_info = demo_study_reflex['trials_sections']
    ts = s2pr.info.TrialsSections(study_ts_info)
    ts.add_subject()
    ts.add_section()
    assert ts.trials_sections == {'biphasic_conv': ['mmax', 'ramp', 'hreflex'],
 'biphasic_high_fq': ['mmax', 'ramp', 'hreflex'],
 'monophasic_conv': ['mmax', 'ramp', 'hreflex'],
 'monophasic_high_fq': ['mmax', 'ramp', 'hreflex']}


def test_get_attributes_complex1(demo_study_reflex):
    study_ts_info = demo_study_reflex['trials_sections']
    ts = s2pr.info.TrialsSections(study_ts_info)
    sub_ts = {'a': ['mmax', 'ramp', 'hreflex'],
                                  'b': ['mmax', 'ramp', 'hreflex'],
                                  'c': ['mmax', 'ramp', 'hreflex'],
                                  'd': ['mmax']}
    ts.add_subject(sub_ts)
    ts.add_section()
    assert ts.trials_sections == sub_ts


def test_get_attributes_complex2(demo_study_reflex):
    study_ts_info = demo_study_reflex['trials_sections']
    ts = s2pr.info.TrialsSections(study_ts_info)

    ts.add_subject()
    sec_ts = {'1': ['mmfax', 'ramhp', 'hrneflex'],
              '2': ['mbmax', 'rnamp', 'hdreflex'],
              '3': ['mmaix', 'raymp', 'hre4flex'],
              '4': ['m5max']}
    ts.add_section(sec_ts)
    assert ts.trials_sections == sec_ts




import numpy as np

import spike2py_reflex as s2pr


def calculate_mean_outcomes(section):
    """Calculate mean outcomes

    For each muscle and for each type of extracted reflex (e.g. H-reflex, mMax),
    calculate the mean outcome. If stimulation was given at different
    intensities, the output is grouped by stim intensity. That is, the mean value
    is computed for all reflexes at a given intensity.

    Example
    -------
    For a given intensity, there are 3 Singles.
    Outcomes such as peak-to-peak have been calculated for each Single, thus we have 3 peak-to-peak values.
    Here, we calculate the mean outcome, which is the average of the 3 peak-to-peak values.

    For Doubles and ratios, each pair of reflexes (reflex 1 and reflex 2) have had their ratio computed.
    For example, lets assume we have 5 Doubles, and we computed the ratio for each Double.
    The mean ratio is the average of the 5 ratios computed for each of the 5 Doubles.

    Returns
    -------
    For single:

    reflexes.mean_outcomes[reflex_type][stim_intensity] = {"outcomes": outcomes,
                                                           "missing_outcomes": missing_outcomes,
                                                           "present_outcomes": present_outcomes}

    For double:

    reflexes.mean_outcomes_reflex1[reflex_type][stim_intensity] = {"outcomes": outcomes1,
                                                                   "missing_outcomes": missing_outcomes1,
                                                                   "present_outcomes": present_outcomes1}
    reflexes.mean_outcomes_reflex2[reflex_type][stim_intensity] = {"outcomes": outcomes2,
                                                                   "missing_outcomes": missing_outcomes2,
                                                                   "present_outcomes": present_outcomes2}
    reflexes.mean_ratio[reflex_type][stim_intensity] = {"ratio": ratio,
                                                        "missing_ratio": ratio_none,
                                                        "present_ratio": ratio_yes}

    """
    for muscle, reflexes in section.reflexes.items():
        stim_intensities = s2pr.outcomes.get_stim_intensities(reflexes)
        if reflexes.type in [s2pr.utils.SINGLE, s2pr.utils.TRAIN]:
            reflexes.mean_outcomes = dict()
            r = _single_mean_outcomes(reflexes, stim_intensities, section.info.section)
            section.reflexes[muscle] = r
        elif reflexes.type == s2pr.utils.DOUBLE:
            reflexes.mean_outcomes_reflex1 = dict()
            reflexes.mean_outcomes_reflex2 = dict()
            reflexes.mean_ratio = dict()
            section.reflexes[muscle] = _double_mean_outcomes(reflexes, stim_intensities, section.info.section)

    return section


def _single_mean_outcomes(reflexes, stim_intensities, section_name):
    if _no_stim_intensities(stim_intensities):
        reflexes = _get_single_mean(reflexes, stim_intensities, section_name)
    else:
        for stim_intensity in stim_intensities:
            reflexes = _get_single_mean(reflexes, stim_intensity, section_name)
    return reflexes


def _no_stim_intensities(stim_intensities):
    return not stim_intensities


def _get_single_mean(reflexes, stim_intensity, section_name):

    # Get idx windows of all reflexes types (e.g. H-reflex + mMax; or just mMax)
    reflex_win_idx_all = reflexes.reflex_windows_idx[section_name]

    # Extract outcomes for each reflex_type of interest (e.g. H-reflex, mMax)
    for reflex_type, _ in reflex_win_idx_all.items():

        if reflex_type not in reflexes.mean_outcomes.keys():
            reflexes.mean_outcomes[reflex_type] = dict()

        # 'yes' variables keep track of how many values used for mean
        # 'none' variables keep track of how many missing values
        peak_to_peak = list()
        peak_to_peak_none = 0
        peak_to_peak_yes = 0
        area = list()
        area_none = 0
        area_yes = 0
        onset = list()
        onset_none = 0
        onset_yes = 0

        # Loop over each possible reflex
        for reflex in reflexes.reflexes:

            # For given intensity, make sure reflex elicited with that intensity
            # If no stim intensity available, include all reflexes
            if ((reflex.stim_intensity == stim_intensity)
                    or (stim_intensity is None)):

                # peak-to-peak
                if reflex.outcomes[reflex_type].peak_to_peak is not None:
                    peak_to_peak.append(reflex.outcomes[reflex_type].peak_to_peak)
                    peak_to_peak_yes += 1
                else:
                    peak_to_peak_none += 1

                # area
                if reflex.outcomes[reflex_type].area is not None:
                    area.append(reflex.outcomes[reflex_type].area)
                    area_yes += 1
                else:
                    area_none += 1

                # onset
                if reflex.outcomes[reflex_type].onset is not None:
                    onset.append(reflex.outcomes[reflex_type].onset)
                    onset_yes += 1
                else:
                    onset_none += 1

        # Calculate mean outcome values

        # peak-to-peak
        if len(peak_to_peak) != 0:
            peak_to_peak = np.mean(peak_to_peak)
        else:
            peak_to_peak = None

        # area
        if len(area) != 0:
            area = np.mean(area)
        else:
            area = None

        # onset
        if len(onset) != 0:
            onset = np.mean(onset)
        else:
            onset = None

        outcomes = s2pr.outcomes.Outcomes(peak_to_peak, area, onset)
        missing_outcomes = s2pr.outcomes.Outcomes(peak_to_peak_none, area_none, onset_none)
        present_outcomes = s2pr.outcomes.Outcomes(peak_to_peak_yes, area_yes, onset_yes)
        if stim_intensity is None:
            reflexes.mean_outcomes[reflex_type][0] = {"outcomes": outcomes,
                                                                        "missing_outcomes": missing_outcomes,
                                                                        "present_outcomes": present_outcomes}
        else:
            reflexes.mean_outcomes[reflex_type][int(stim_intensity)] = {"outcomes": outcomes,
                                                                        "missing_outcomes": missing_outcomes,
                                                                        "present_outcomes": present_outcomes}
    return reflexes


def _double_mean_outcomes(reflexes, stim_intensities, section_name):

    if _no_stim_intensities(stim_intensities):
        reflexes = _get_double_mean(reflexes, stim_intensities, section_name)
    else:
        for stim_intensity in stim_intensities:
            reflexes = _get_double_mean(reflexes, stim_intensity, section_name)
    return reflexes


def _get_double_mean(reflexes, stim_intensity, section_name):

    # Get idx windows of all reflexes types (e.g. H-reflex + mMax; or just mMax)
    reflex_win_idx_all = reflexes.reflex_windows_idx[section_name]

    # Extract outcomes for each reflex_type of interest (e.g. hreflex, mmax)
    for reflex_type, _ in reflex_win_idx_all.items():

        if reflex_type not in reflexes.mean_outcomes_reflex1.keys():
            reflexes.mean_ratio[reflex_type] = dict()
            reflexes.mean_outcomes_reflex1[reflex_type] = dict()
            reflexes.mean_outcomes_reflex2[reflex_type] = dict()

        # 'yes' variables keep track of how many values used for mean
        # 'none' variables keep track of how many missing values
        peak_to_peak1 = list()
        peak_to_peak_none1 = 0
        peak_to_peak_yes1 = 0
        area1 = list()
        area_none1 = 0
        area_yes1 = 0
        onset1 = list()
        onset_none1 = 0
        onset_yes1 = 0

        peak_to_peak2 = list()
        peak_to_peak_none2 = 0
        peak_to_peak_yes2 = 0
        area2 = list()
        area_none2 = 0
        area_yes2 = 0
        onset2 = list()
        onset_none2 = 0
        onset_yes2 = 0

        ratio = list()
        ratio_none = 0
        ratio_yes = 0

        for reflex in reflexes.reflexes:

            # For given intensity, make sure reflex elicited with that intensity
            # If no stim intensity available, include all reflexes
            if ((reflex.stim_intensity == stim_intensity)
                    or (stim_intensity is None)):

                # peak-to-peak
                if reflex.reflex1.outcomes[reflex_type].peak_to_peak is not None:
                    peak_to_peak1.append(reflex.reflex1.outcomes[reflex_type].peak_to_peak)
                    peak_to_peak_yes1 += 1
                else:
                    peak_to_peak_none1 += 1

                # area
                if reflex.reflex1.outcomes[reflex_type].area is not None:
                    area1.append(reflex.reflex1.outcomes[reflex_type].area)
                    area_yes1 += 1
                else:
                    area_none1 += 1

                # onset
                if reflex.reflex1.outcomes[reflex_type].onset is not None:
                    onset1.append(reflex.reflex1.outcomes[reflex_type].onset)
                    onset_yes1 += 1
                else:
                    onset_none1 += 1

                # peak-to-peak
                try:
                    if reflex.reflex2.outcomes[reflex_type].peak_to_peak is not None:
                        peak_to_peak2.append(reflex.reflex2.outcomes[reflex_type].peak_to_peak)
                        peak_to_peak_yes2 += 1
                    else:
                        peak_to_peak_none2 += 1

                    # area
                    if reflex.reflex2.outcomes[reflex_type].area is not None:
                        area2.append(reflex.reflex2.outcomes[reflex_type].area)
                        area_yes2 += 1
                    else:
                        area_none2 += 1

                    # onset
                    if reflex.reflex2.outcomes[reflex_type].onset is not None:
                        onset2.append(reflex.reflex2.outcomes[reflex_type].onset)
                        onset_yes2 += 1
                    else:
                        onset_none2 += 1

                except AttributeError:
                    pass

                try:
                    # ratio
                    if reflex.ratio[reflex_type] is not None:
                        ratio.append(reflex.ratio[reflex_type])
                        ratio_yes += 1
                    else:
                        ratio_none += 1
                except KeyError:
                    pass

        # Calculate mean outcome values

        # peak-to-peak1
        if len(peak_to_peak1) != 0:
            peak_to_peak1 = np.mean(peak_to_peak1)
        else:
            peak_to_peak1 = None

        # area1
        if len(area1) != 0:
            area1 = np.mean(area1)
        else:
            area1 = None

        # onset1
        if len(onset1) != 0:
            onset1 = np.mean(onset1)
        else:
            onset1 = None

        # peak-to-peak2
        if len(peak_to_peak2) != 0:
            peak_to_peak2 = np.mean(peak_to_peak2)
        else:
            peak_to_peak2 = None
        # area2
        if len(area2) != 0:
            area2 = np.mean(area2)
        else:
            area2 = None
        # onset2
        if len(onset2) != 0:
            onset2 = np.mean(onset2)
        else:
            onset2 = None

        # ratio
        if len(ratio) != 0:
            ratio = np.mean(ratio)
        else:
            ratio = None

        outcomes1 = s2pr.outcomes.Outcomes(peak_to_peak1, area1, onset1)
        missing_outcomes1 = s2pr.outcomes.Outcomes(peak_to_peak_none1, area_none1, onset_none1)
        present_outcomes1 = s2pr.outcomes.Outcomes(peak_to_peak_yes1, area_yes1, onset_yes1)

        outcomes2 = s2pr.outcomes.Outcomes(peak_to_peak2, area2, onset2)
        missing_outcomes2 = s2pr.outcomes.Outcomes(peak_to_peak_none2, area_none2, onset_none2)
        present_outcomes2 = s2pr.outcomes.Outcomes(peak_to_peak_yes2, area_yes2, onset_yes2)

        reflexes.mean_outcomes_reflex1[reflex_type][int(stim_intensity)] = {"outcomes": outcomes1,
                                                                            "missing_outcomes": missing_outcomes1,
                                                                            "present_outcomes": present_outcomes1}
        reflexes.mean_outcomes_reflex2[reflex_type][int(stim_intensity)] = {"outcomes": outcomes2,
                                                                            "missing_outcomes": missing_outcomes2,
                                                                            "present_outcomes": present_outcomes2}
        reflexes.mean_ratio[reflex_type][int(stim_intensity)] = {"ratio": ratio,
                                                                 "missing_ratio": ratio_none,
                                                                 "present_ratio": ratio_yes}

    return reflexes



from .reflexes import reflexes
from .single import (plot_single,
                     get_muscles_and_intensities,
                     get_plot_width_height,
                     overlay_reflexes,
                     set_x_lim,
                     save_figure,
                     add_reflex_windows)
from .double import plot_double
from .train import plot_train

from .outcomes import single_outcomes, train_outcomes, double_outcomes
from .outcomes import plot_outcomes as outcomes

from .plots import subject as summary_subject
from .plots import study as summary_study



import copy


class TrialsSections:
    def __init__(self, study_trials_sections=None):
        self._study = None
        self._subject = None
        self._section = None
        if study_trials_sections is not None:
            self._study = study_trials_sections

    def add_subject(self, subject_trials_sections=None):
        if subject_trials_sections is not None:
            self._subject = subject_trials_sections
        elif subject_trials_sections is None:
            self._subject = copy.deepcopy(self._study)

    def add_section(self, section_trials_sections=None):
        if section_trials_sections is not None:
            self._section = section_trials_sections
        elif section_trials_sections is None:
            self._section = copy.deepcopy(self._subject)

    def clear_subject(self):
        self._subject = None

    def clear_section(self):
        self._section = None

    @property
    def trials_sections(self):
        if self._section is not None:
            return self._section
        else:
            return self._subject



import spike2py_reflex as s2pr


def test_windows_types():
    window = s2pr.info.Window(single=1,
                            double=2,
                            double_single_pulse=3,
                            train_single_pulse=4)
    window_types = s2pr.info.WindowTypes(extract=window,
                                       plotting=window,
                                       reflexes={'a': 5},
                                       sd=[6, 7])
    assert window_types.extract.single == 1
    assert window_types.reflexes['a'] == 5

    window_types.clear()

    assert window_types.extract.single is None
    assert window_types.reflexes is None


def test_grouped_windows_initialise_with_win_info(demo_study_reflex):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.GroupedWindows(win_info=win_info)

    assert win.ms.extract.single == [-250, 100]
    assert win.ms.plotting.double == [-25, 75]
    assert win.ms.sd == [-205, -5]

    assert win.idx.extract.single is None
    assert win.idx.plotting.double is None
    assert win.idx.sd is None


def test_grouped_windows_update(demo_study_reflex):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.GroupedWindows(win_info=win_info)
    assert win.ms.sd == [-205, -5]

    new_window = [-50, -3]
    win.update({"sd": new_window})
    assert win.ms.sd == new_window

    extract_data = {"single": [1, 1],
                    "double": [2, 2],
                    "double_single_pulse": [3, 3],
                    "train_single_pulse": [4, 4]}
    win.update({"extract": extract_data})
    assert win.ms.extract.single == [1, 1]


def test_grouped_windows_update_fs(demo_study_reflex):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.GroupedWindows(win_info=win_info)
    win.add_double_isi(50)
    win.add_fs(2000)


def test_grouped_windows_compute_x_axis(demo_study_reflex):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.GroupedWindows(win_info=win_info)
    assert win._x_axis is None
    win.add_double_isi(50)
    win.fs = 1984
    win._get_x_axis()
    assert len(win._x_axis) == 3967


def test_grouped_windows_compute_idx(demo_study_reflex):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.GroupedWindows(win_info=win_info)
    win.add_double_isi(50)
    win.add_fs(1984)
    assert win.idx.sd['single'] == [89, 486]
    assert win.idx.reflexes['Fdi']['mmax'] == {'hreflex': {'double': [[536, 575], [635, 675]],
             'single': [536, 575],
             'train': [70, 89]},
 'mmax': {'double': [[512, 536], [611, 635]],
          'single': [512, 536],
          'train': [46, 70]}}
    assert win.idx.extract.double == [-496, 298]


def test_grouped_windows_compute_x_axes(demo_study_reflex):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.GroupedWindows(win_info=win_info)
    win.add_double_isi(50)
    win.add_fs(1984)
    assert len(win.x_axes.single) == 694
    assert len(win.x_axes.double) == 794
    assert len(win.x_axes.double_single_pulse) == 109
    assert len(win.x_axes.train_single_pulse) == 90


def test_windows_init(demo_study_reflex):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.Windows(win_info)
    assert win._study.ms.sd == [-205, -5]


def test_windows_add_fs_compute_idx(demo_study_reflex):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.Windows(win_info)
    win.double_isi = 50
    win.fs = 1984
    assert win._study.idx.sd == {'double': [89, 486], 'single': [89, 486]}


def test_windows_add_subject_long(demo_study_reflex, demo_subject_windows_long):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.Windows(win_info)
    win.add_subject(demo_subject_windows_long)
    assert win._study.ms.sd == [-205, -5]
    assert win._subject.ms.sd == [-50, -20]
    assert win._subject.ms.plotting.single == [-25, 50]
    win.clear_subject()
    assert win._subject.ms.sd == None


def test_windows_add_subject_short(demo_study_reflex, demo_subject_windows_short):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.Windows(win_info)
    win.add_subject(demo_subject_windows_short)

    assert win._study.ms.sd == [-205, -5]
    assert win._subject.ms.sd == [-51, -21]

    assert win._study.ms.reflexes['Fdi']['mmax'] == {'hreflex': [20, 40], 'mmax': [8, 20]}
    assert win._subject.ms.reflexes['Fdi']['mmax'] == {'hreflex': [26, 45], 'mmax': [6, 21]}

    assert win._study.ms.extract.single == [-250, 100]
    assert win._subject.ms.extract.single == [-250, 100]


def test_windows_add_subject_section(demo_study_reflex, demo_subject_windows_short, demo_section_windows_short):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.Windows(win_info)
    win.add_subject(demo_subject_windows_short)
    win.add_section(demo_section_windows_short)

    assert win._study.ms.sd == [-205, -5]
    assert win._subject.ms.sd == [-51, -21]
    assert win._section.ms.sd == [-58, -26]

    assert win._study.ms.reflexes['Fdi']['mmax'] == {'hreflex': [20, 40], 'mmax': [8, 20]}
    assert win._subject.ms.reflexes['Fdi']['mmax'] == {'hreflex': [26, 45], 'mmax': [6, 21]}
    assert win._section.ms.reflexes['Fdi']['mmax'] == {'hreflex': [19, 39], 'mmax': [8, 26]}

    assert win._study.ms.extract.single == [-250, 100]
    assert win._subject.ms.extract.single == [-250, 100]
    assert win._section.ms.extract.single == [-250, 100]


def test_windows_add_subject_section_get_idx(demo_study_reflex, demo_subject_windows_short, demo_section_windows_short):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.Windows(win_info)
    win.add_subject(demo_subject_windows_short)
    win.add_section(demo_section_windows_short)
    win.double_isi = 50
    win.fs = 1984
    assert win._study.idx.sd == {'double': [89, 486], 'single': [89, 486]}
    assert win._subject.idx.sd == {'double': [395, 454], 'single': [395, 454]}
    assert win._section.idx.sd == {'double': [381, 444], 'single': [381, 444]}

    assert win._study.idx.reflexes['Fdi']['mmax'] == {'hreflex': {'double': [[536, 575], [635, 675]],
             'single': [536, 575],
             'train': [70, 89]},
 'mmax': {'double': [[512, 536], [611, 635]],
          'single': [512, 536],
          'train': [46, 70]}}
    assert win._subject.idx.reflexes['Fdi']['mmax'] == {'hreflex': {'double': [[548, 585], [647, 684]],
             'single': [548, 585],
             'train': [82, 89]},
 'mmax': {'double': [[508, 538], [607, 637]],
          'single': [508, 538],
          'train': [42, 72]}}
    assert win._section.idx.reflexes['Fdi']['mmax'] == {'hreflex': {'double': [[534, 573], [633, 673]],
             'single': [534, 573],
             'train': [68, 89]},
 'mmax': {'double': [[512, 548], [611, 647]],
          'single': [512, 548],
          'train': [46, 82]}}

    assert win._study.idx.extract.single == [-496, 198]
    assert win._subject.idx.extract.single == [-496, 198]
    assert win._section.idx.extract.single == [-496, 198]


def test_windows_get_attributes(demo_study_reflex, demo_subject_windows_short, demo_section_windows_short):
    win_info = demo_study_reflex['windows']
    win = s2pr.info.Windows(win_info)
    win.add_subject(demo_subject_windows_short)
    win.add_section(demo_section_windows_short)
    win.double_isi = 50
    win.fs = 1984

    assert win.idx_sd == {'double': [381, 444], 'single': [381, 444]}
    assert win.idx_reflexes['Fdi']['mmax'] == {'hreflex': {'double': [[534, 573], [633, 673]],
             'single': [534, 573],
             'train': [68, 89]},
 'mmax': {'double': [[512, 548], [611, 647]],
          'single': [512, 548],
          'train': [46, 82]}}
    assert win.idx_extract.single == [-496, 198]

# # TODO: Add a test to make sure that if sd window not specified, sd idx set to None



from pathlib import Path

import typer

import process as study_


app = typer.Typer(name="spike2py_reflex", help="Extract evoked responses from Spike2 data.")


@app.command()
def section(
    subject_id: str, study_path: str, trial_name: str, section_name: str, plot: bool = False
):
    """Extract reflexes from a given a section.

    Requires that the proper study file-folder structure be in place
    and that the study_info.json and study_reflex.json be present
    in the `study_path`. Note that a subject_reflex.json file is present
    in the subject folder it will be used and override study-level
    parameters. Similarly, if a section-specific json file is present in the
    subject's `proc` folder, it will be used and override study-level and
    subject-level parameters (e.g. <trial_name>_<section_name>_reflex.json).

    trial_info_json: Path to json file containing details required by spike2py.trial.TrialInfo

    Parameters
    ----------
    subject_id: Subject id that matches the folder name for the study.
    study_path: path to study folder.
    trial_name: Name of the trial containing section to be analysed
    section_name: Name of the section to analyssed.
    plot: Flag to indicate whether to generate and save figures.

    """
    reflex_study_info = study_.get_reflex_study_info(study_path)
    reflex_subject_info = subject_.get_reflex_subject_info(subject_id, reflex_study_info)
    section_.process(reflex_subject_info, trial_name, section_name, plot)


@app.command()
def subject(subject_id: str, study_path: str, plot: bool = False):
    """Extract reflexes from all sections of each trials for a given subject.

    Requires that the proper study file-folder structure be in place
    and that the study_info.json and study_reflex.json be present
    in the `study_path`. Also, a subject_reflex.json file can be present
    to override study-specific details.

    Parameters
    ----------
    subject_id: Subject id that matches the folder name for the study.
    study_path: path to study folder.
    plot: Flag to indicate whether to generate and save figures.
    """
    pass

@app.command()
def study(study_path: str, plot: bool = False):
    """Extract reflexes from all sections of each trials from all subjects for a study.

    Parameters
    ----------
    study_path: Path to study folder.
    plot: Flag to indicate whether to generate and save figures.
    """
    study_.process(Path(study_path), plot=plot)


if __name__ == "__main__":
    app()



CONVERT_MS_TO_S = 0.001
KHZ_ISI_THRESHOLD = 0.0011  # 1Kz and above will be removed

SINGLE = "single"
DOUBLE = "double"
TRAIN = "train"

SD_MULTIPLIER = 3

TRAIN_SINGLE_PULSE = "train_single_pulse"
STUDY_INFO_FILE = "study_info.json"
REFLEX_FILE = "study_reflex.json"
SUBJECT_REFLEX_FILE = "subject_reflex.json"



from pathlib import Path
import pickle
import spike2py_preprocess as s2pp
import spike2py_reflex as s2pr


def study(study_path: Path):
    """Extract reflexes for all trials of all subject for a study

    Parameters
    ----------
    study_path: Path to study folder. Folder is to be structured as indicated in spike2py_preprocess docs
    """
    info = s2pr.info.Info(study_path)
    for subject_ in info.subjects:
        subject(subject_, study_path=study_path)


def subject(subject_: str, study_path: Path):
    """Process all trials and their sections for a given subject

    Parameters
    ----------
    subject_: Subject id, which also has to be the name of their folder
    study_path: Path to study folder
    info: Study-related details; required if called internally
    plot: Whether to generate plots

    """
    print(f"\tsubject: {subject_}")
    subject_folder = study_path / subject_ / "proc"
    for item in subject_folder.iterdir():
        if item.is_file():
            if item.stem.split("_")[-1] == "reflexes":
                print(f"\t\t plotting {item.stem}")
                with open(item, "rb") as f:
                    section = pickle.load(f)

                s2pr.plot.reflexes(section)
                s2pr.plot.outcomes(section)

    subject_path_reflexes = study_path / subject_ / "figures" / "reflexes"
    subject_path_outcomes = study_path / subject_ / "figures" / "outcomes"

    # Delete previously concatenated figure file if present
    subject_path_reflexes_file = subject_path_reflexes / (subject_ + '.pdf')
    if subject_path_reflexes_file.is_file():
        subject_path_reflexes_file.unlink()

    subject_path_outcomes_file = subject_path_reflexes / (subject_ + '.pdf')
    if subject_path_outcomes_file.is_file():
        subject_path_outcomes.unlink()

    s2pp.utils.merge_pdfs(subject_path_reflexes)
    s2pp.utils.merge_pdfs(subject_path_outcomes)



import spike2py_reflex as s2pr


LINE = 80 * '-'
TRIGGER = 'trig'
MUSCLE = 'muscle'
REFLEX = 'reflex'
STIM = 'stim'
TIME1 = 'time1'
TIME2 = 'time2'
P2P = 'p2p'
AREA = 'area'
ONSET = 'onset'


def summary(section):
    muscles = list(section.reflexes.keys())
    if section.reflexes[muscles[0]].type == s2pr.utils.SINGLE:
        _single_summary(section)
    elif section.reflexes[muscles[0]].type == s2pr.utils.TRAIN:
        _train_summary(section)
    elif section.reflexes[muscles[0]].type == s2pr.utils.DOUBLE:
        _double_summary(section)


def _single_summary(section):
    summary = list()
    line = f'{section.info.trial.upper()} | {section.info.section.upper()}'
    summary = _add_header(summary, line)

    muscles, _, intensities, _ = s2pr.plot.get_muscles_and_intensities(section)
    if intensities[0] == 'no_intensity':
        intensities = [0]
    section_name = section.info.section
    reflex_types = list(section.info.windows.ms_reflexes[muscles[0]][section_name].keys())

    line = (f'{TRIGGER:<5s}{MUSCLE:<8s}{REFLEX:<10s}{STIM:<8s}{TIME1:<10s}'
            f'{TIME2:<10s}{P2P:^8s}  {AREA:^8s}  {ONSET:^8s}')

    summary = _add_header(summary, line)

    for muscle in muscles:
        for reflex_type in reflex_types:
            for i, reflex in enumerate(section.reflexes[muscle].reflexes):
                summary = _add_single_reflex_outcomes(i, reflex, reflex_type, muscle, summary)
            summary.append(LINE)
    summary = _add_mean_single_reflex_outcomes(summary, section, muscles, intensities, reflex_types)
    _save_summary(summary, section)


def _add_header( summary, line):
    summary.append(LINE)
    summary.append(line)
    summary.append(LINE)
    return summary


def _add_single_reflex_outcomes(i, reflex, reflex_type, muscle, summary):
    if reflex.outcomes[reflex_type].onset is None:
        line = (f'{i:<5d}{muscle:<8s}{reflex_type:<10s}{reflex.stim_intensity:<8d}'
                f'{reflex.extract_times[0]:<10.2f}{reflex.extract_times[1]:<10.2f}'
                f'{reflex.outcomes[reflex_type].peak_to_peak:>8.4f}'
                f'  {reflex.outcomes[reflex_type].area:>8.4f}')
    else:
        line = (f'{i:<5d}{muscle:<8s}{reflex_type:<10s}{reflex.stim_intensity:<8d}'
                f'{reflex.extract_times[0]:<10.2f}{reflex.extract_times[1]:<10.2f}'
                f'{reflex.outcomes[reflex_type].peak_to_peak:>8.4f}'
                f'  {reflex.outcomes[reflex_type].area:>8.4f}'
                f'  {reflex.outcomes[reflex_type].onset:>8.4f}')
    summary.append(line)
    return summary


def _add_mean_single_reflex_outcomes(summary, section, muscles, intensities, reflex_types):
    if intensities is None:
        intensities = [0]
    i = ''
    summary.append('MEAN VALUES')
    summary.append(LINE)
    for muscle in muscles:
        for reflex_type in reflex_types:
            for intensity in intensities:
                p2p = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['outcomes'].peak_to_peak
                area = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['outcomes'].area
                onset = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['outcomes'].onset
                if onset is None:
                    line = (f'{i:<5s}{muscle:<8s}{reflex_type:<10s}{intensity:<28d}'
                            f'{p2p:>8.4f}'
                            f'  {area:>8.4f}')
                else:

                    line = (f'{i:<5s}{muscle:<8s}{reflex_type:<10s}{intensity:<28d}'
                            f'{p2p:>8.4f}'
                            f'  {area:>8.4f}'
                            f'  {onset:>8.4f}')
                summary.append(line)

                p2p_missing = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['missing_outcomes'].peak_to_peak
                p2p_present = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['present_outcomes'].peak_to_peak
                p2p_val = str(p2p_present) + '/' + str(p2p_present + p2p_missing)

                area_missing = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['missing_outcomes'].area
                area_present = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['present_outcomes'].area
                area_val = str(area_present) + '/' + str(area_present + area_missing)

                if onset is None:
                    onset_val = ""
                else:
                    onset_missing = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['missing_outcomes'].onset
                    onset_present = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['present_outcomes'].onset
                    onset_val = str(onset_present) + '/' + str(onset_present + onset_missing)
                line = (f'{p2p_val:>56s}'
                        f'  {area_val:>8s}'
                        f'  {onset_val:>8s}')
                summary.append(line)
                summary.append(LINE)
    return summary


def _save_summary(summary, section):
    file_name = f'{section.info.trial}_{section.info.section}_summary.txt'
    path = section.info.study_path / section.info.subject / 'numerical_outcomes' / 'summary'
    if not path.exists():
        path.mkdir(parents=True)
    with open(path / file_name, 'w') as f:
        for line in summary:
            f.write(line + '\n')


def _double_summary(section):
    summary = list()
    line = f'{section.info.trial.upper()} | {section.info.section.upper()}'
    summary = _add_header(summary, line)

    muscles, _, intensities, _ = s2pr.plot.get_muscles_and_intensities(section)
    section_name = section.info.section
    reflex_types = list(section.info.windows.ms_reflexes[muscles[0]][section_name].keys())

    line = (f'{TRIGGER:<5s}{MUSCLE:<8s}{REFLEX:<10s}{STIM:<8s}{TIME1:<10s}'
            f'{TIME2:<10s}{P2P:^8s}  {AREA:^8s}  {ONSET:^8s}')

    summary = _add_header(summary, line)

    for muscle in muscles:
        for reflex_type in reflex_types:
            for i, reflex in enumerate(section.reflexes[muscle].reflexes):
                summary = _add_double_reflex_outcomes(i, reflex, reflex_type, muscle, summary)
            summary.append(LINE)
    summary = _add_mean_double_reflex_outcomes(summary, section, muscles, intensities, reflex_types)
    _save_summary(summary, section)


def _add_double_reflex_outcomes(i, reflex, reflex_type, muscle, summary):
    i1 = str(i) + '-1'
    i2 = str(i) + '-2'

    try:
        if reflex.reflex1.outcomes[reflex_type].onset is None:
            line1 = (f'{i1:<5s}{muscle:<8s}{reflex_type:<10s}{reflex.stim_intensity:<8d}'
                    f'{reflex.extract_times[0]:<10.2f}{reflex.extract_times[1]:<10.2f}'
                    f'{reflex.reflex1.outcomes[reflex_type].peak_to_peak:>8.4f}'
                    f'  {reflex.reflex1.outcomes[reflex_type].area:>8.4f}')
        else:
            line1 = (f'{i1:<5s}{muscle:<8s}{reflex_type:<10s}{reflex.stim_intensity:<8d}'
                    f'{reflex.extract_times[0]:<10.2f}{reflex.extract_times[1]:<10.2f}'
                    f'{reflex.reflex1.outcomes[reflex_type].peak_to_peak:>8.4f}'
                    f'  {reflex.reflex1.outcomes[reflex_type].area:>8.4f}'
                    f'  {reflex.reflex1.outcomes[reflex_type].onset:>8.4f}')
        summary.append(line1)
    except AttributeError:
        print('Summary outcomes: reflex1 missing')
    try:
        if reflex.reflex2.outcomes[reflex_type].onset is None:
            line2 = (f'{i2:<51s}'
                    f'{reflex.reflex2.outcomes[reflex_type].peak_to_peak:>8.4f}'
                    f'  {reflex.reflex2.outcomes[reflex_type].area:>8.4f}')
        else:
            line2 = (f'{i2:<51s}'
                     f'{reflex.reflex2.outcomes[reflex_type].peak_to_peak:>8.4f}'
                     f'  {reflex.reflex2.outcomes[reflex_type].area:>8.4f}'
                     f'  {reflex.reflex2.outcomes[reflex_type].onset:>8.4f}')
        summary.append(line2)
    except AttributeError:
        print('Summary outcomes: reflex1 missing')
    return summary


def _add_mean_double_reflex_outcomes(summary, section, muscles, intensities, reflex_types):
    if intensities is None:
        intensities = [0]
    i1 = '1'
    i2 = '2'
    summary.append('MEAN VALUES')
    summary.append(LINE)
    for muscle in muscles:
        for reflex_type in reflex_types:
            for intensity in intensities:
                p2p1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['outcomes'].peak_to_peak
                area1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['outcomes'].area
                onset1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['outcomes'].onset
                try:
                    p2p2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['outcomes'].peak_to_peak
                    area2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['outcomes'].area
                    onset2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['outcomes'].onset
                except AttributeError:
                    p2p2 = None
                    area2 = None
                    onset2 = None

                p2p_missing1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['missing_outcomes'].peak_to_peak
                p2p_present1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['present_outcomes'].peak_to_peak
                p2p_val1 = str(p2p_present1) + '/' + str(p2p_present1 + p2p_missing1)

                area_missing1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['missing_outcomes'].area
                area_present1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['present_outcomes'].area
                area_val1 = str(area_present1) + '/' + str(area_present1 + area_missing1)

                try:
                    p2p_missing2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['missing_outcomes'].peak_to_peak
                    p2p_present2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['present_outcomes'].peak_to_peak
                    p2p_val2 = str(p2p_present2) + '/' + str(p2p_present2 + p2p_missing2)

                    area_missing2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['missing_outcomes'].area
                    area_present2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['present_outcomes'].area
                    area_val2 = str(area_present2) + '/' + str(area_present2 + area_missing2)
                except AttributeError:
                    p2p_missing2 = None
                    p2p_present2 = None
                    p2p_val2 = None
                    area_missing2 = None
                    area_present2 = None
                    area_val2 = None

                if onset1 is None:
                    line1 = (f'{i1:<5s}{muscle:<8s}{reflex_type:<10s}{intensity:<28d}'
                            f'{p2p1:>8.4f}'
                            f'  {area1:>8.4f}')
                    onset_val1 = ""
                else:
                    line1 = (f'{i1:<5s}{muscle:<8s}{reflex_type:<10s}{intensity:<28d}'
                            f'{p2p1:>8.4f}'
                            f'  {area1:>8.4f}'
                            f'  {onset1:>8.4f}')
                    onset_missing1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['missing_outcomes'].onset
                    onset_present1 = section.reflexes[muscle].mean_outcomes_reflex1[reflex_type][intensity]['present_outcomes'].onset
                    onset_val1 = str(onset_present1) + '/' + str(onset_present1 + onset_missing1)

                line11 = (f'{p2p_val1:>56s}'
                          f'  {area_val1:>8s}'
                          f'  {onset_val1:>8s}')

                summary.append(line1)
                summary.append(line11)

                if p2p2 is not None:
                    if onset2 is None:
                        line2 = (f'{i2:<51s}'
                                f'{p2p2:>8.4f}'
                                f'  {area2:>8.4f}')
                        onset_val2 = ""
                    else:
                        line2 = (f'{i2:<51s}'
                                f'{p2p2:>8.4f}'
                                f'  {area2:>8.4f}'
                                f'  {onset2:>8.4f}')

                        onset_missing2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['missing_outcomes'].onset
                        onset_present2 = section.reflexes[muscle].mean_outcomes_reflex2[reflex_type][intensity]['present_outcomes'].onset
                        onset_val2 = str(onset_present2) + '/' + str(onset_present2 + onset_missing2)

                    line22 = (f'{p2p_val2:>56s}'
                              f'  {area_val2:>8s}'
                              f'  {onset_val2:>8s}')
                    summary.append(line2)
                    summary.append(line22)
    return summary


def _train_summary(section):
    summary = list()
    line = f'{section.info.trial.upper()} | {section.info.section.upper()}'
    summary = _add_header(summary, line)

    muscles, _, intensities, _ = s2pr.plot.get_muscles_and_intensities(section)
    section_name = section.info.section
    reflex_types = list(section.info.windows.ms_reflexes[muscles[0]][section_name].keys())

    line = (f'{TRIGGER:<5s}{MUSCLE:<8s}{REFLEX:<10s}{STIM:<8s}{TIME1:<10s}'
            f'{TIME2:<10s}{P2P:^8s}  {AREA:^8s}')

    summary = _add_header(summary, line)

    for muscle in muscles:
        for reflex_type in reflex_types:
            for i, reflex in enumerate(section.reflexes[muscle].reflexes):
                summary = _add_train_reflex_outcomes(i, reflex, reflex_type, muscle, summary)
            summary.append(LINE)
    summary = _add_mean_train_reflex_outcomes(summary, section, muscles, intensities, reflex_types)
    _save_summary(summary, section)


def _add_train_reflex_outcomes(i, reflex, reflex_type, muscle, summary):
    line = (f'{i:<5d}{muscle:<8s}{reflex_type:<10s}{reflex.stim_intensity:<8d}'
            f'{reflex.extract_times[0]:<10.2f}{reflex.extract_times[1]:<10.2f}'
            f'{reflex.outcomes[reflex_type].peak_to_peak:>8.4f}'
            f'  {reflex.outcomes[reflex_type].area:>8.4f}')
    summary.append(line)
    return summary


def _add_mean_train_reflex_outcomes(summary, section, muscles, intensities, reflex_types):
    if intensities is None:
        intensities = [0]
    i = ''
    summary.append('MEAN VALUES')
    summary.append(LINE)
    for muscle in muscles:
        for reflex_type in reflex_types:
            for intensity in intensities:
                p2p = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['outcomes'].peak_to_peak
                area = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['outcomes'].area
                line = (f'{i:<5s}{muscle:<8s}{reflex_type:<10s}{intensity:<28d}'
                        f'{p2p:>8.4f}'
                        f'  {area:>8.4f}')

                summary.append(line)

                p2p_missing = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['missing_outcomes'].peak_to_peak
                p2p_present = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['present_outcomes'].peak_to_peak
                p2p_val = str(p2p_present) + '/' + str(p2p_present + p2p_missing)

                area_missing = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['missing_outcomes'].area
                area_present = section.reflexes[muscle].mean_outcomes[reflex_type][intensity]['present_outcomes'].area
                area_val = str(area_present) + '/' + str(area_present + area_missing)

                line = (f'{p2p_val:>59s}'
                        f'  {area_val:>7s}')

                summary.append(line)
                summary.append(LINE)
    return summary



import numpy as np

import spike2py_reflex as s2pr


def calculate(section):
    """Calculate outcomes of individual reflexes and all section reflexes

    For the summary outcomes, we calculate:
    1) The mean value of the outcomes from each reflex (or pairs of reflexes),
    2) Outcome value of the mean waveform.

    """
    section = calculate_for_individual_reflexes(section)
    section = s2pr.outcomes.calculate_mean_outcomes(section)
    section = s2pr.outcomes.calculate_outcomes_of_avg(section)
    return section


def calculate_for_individual_reflexes(section):
    """Calculate outcomes for each reflex of each muscle"""
    for muscle, reflexes in section.reflexes.items():
        section_name = section.info.section
        if reflexes.type in [s2pr.utils.SINGLE, s2pr.utils.TRAIN]:
            section.reflexes[muscle] = _get_single_outcomes(reflexes, section_name)
        elif reflexes.type == s2pr.utils.DOUBLE:
            double_isi = section.info.windows.double_isi
            section.reflexes[muscle] = _get_double_outcomes(reflexes, section_name, double_isi)
    return section


def _get_single_outcomes(reflexes, section_name):
    x_axis = reflexes.x_axis_extract
    sd_idx_all_stim_times = reflexes.sd_window_idx
    # Get idx windows of all reflexes types (e.g. H-reflex + mMax; or just mMax)
    reflex_win_idx_all = reflexes.reflex_windows_idx[section_name]

    for i in range(len(reflexes.reflexes)):
        waveform = reflexes.reflexes[i].waveform
        reflexes.reflexes[i].outcomes = dict()
        for reflex_type, reflex_idx_dict in reflex_win_idx_all.items():
            # Extract specific idx windows for current reflex type
            reflex_win_idx = reflex_idx_dict[reflexes.type]
            if reflexes.type == s2pr.utils.SINGLE:
                sd_idx = sd_idx_all_stim_times[reflexes.type]
            else:
                sd_idx = None  # train stimulation
            outcomes, background_sd = get_outcomes_from(waveform, reflex_win_idx, sd_idx, x_axis)
            reflexes.reflexes[i].background_sd = background_sd
            reflexes.reflexes[i].outcomes[reflex_type] = outcomes
    return reflexes


def _get_double_outcomes(reflexes, section_name, double_isi):
    x_axis = reflexes.x_axis_extract
    sd_idx_all = reflexes.sd_window_idx
    reflex_win_idx_all = reflexes.reflex_windows_idx[section_name]

    for i in range(len(reflexes.reflexes)):
        reflexes.reflexes[i].reflex1.outcomes = dict()
        try:
            reflexes.reflexes[i].reflex2.outcomes = dict()
        except AttributeError:
            pass
        reflexes.reflexes[i].ratio = dict()

        waveform = reflexes.reflexes[i].waveform

        for reflex_type, reflex_idx_dict in reflex_win_idx_all.items():
            reflex_win_idx = reflex_idx_dict[reflexes.type]
            sd_idx = sd_idx_all[reflexes.type]

            # Reflex1
            reflex1_idx = reflex_win_idx[0]
            outcomes1, background_sd = get_outcomes_from(waveform, reflex1_idx, sd_idx, x_axis)
            reflexes.reflexes[i].reflex1.background_sd = background_sd
            reflexes.reflexes[i].reflex1.outcomes[reflex_type] = outcomes1

            try:
                # Reflex2
                if reflex_win_idx[1] is not None:
                    reflex2_idx = reflex_win_idx[1]
                    outcomes2, background_sd = get_outcomes_from(waveform, reflex2_idx, sd_idx, x_axis)
                    if outcomes2.onset is not None:
                        outcomes2.onset = outcomes2.onset - (double_isi*s2pr.utils.CONVERT_MS_TO_S)
                    reflexes.reflexes[i].reflex2.background_sd = background_sd
                    reflexes.reflexes[i].reflex2.outcomes[reflex_type] = outcomes2
                    if outcomes2.peak_to_peak is not None:
                        ratio = outcomes2.peak_to_peak / outcomes1.peak_to_peak
                    else:
                        ratio = None
                    reflexes.reflexes[i].ratio[reflex_type] = ratio
            except AttributeError:
                pass
    return reflexes


def get_outcomes_from(waveform, reflex_idx, sd_win_idx, x_axis):
    onset = None
    background_sd = None
    amplitude = _get_amplitude(waveform, reflex_idx[0], reflex_idx[1])
    area = _get_area(waveform, reflex_idx[0], reflex_idx[1])
    if sd_win_idx is not None:
        onset_idx, background_sd = _get_onset(waveform, reflex_idx, sd_win_idx)
        if onset_idx is not None:
            onset = x_axis[onset_idx]
    outcomes = s2pr.outcomes.Outcomes(amplitude, area, onset)
    return outcomes, background_sd


# TODO: Add test to make sure reflex amplitude calc works for various combinations (signal all positive, all negative, mix)
def _get_amplitude(waveform, idx1, idx2):
    min_val = np.min(waveform[idx1:idx2])
    max_val = np.max(waveform[idx1:idx2])
    amplitude = max_val - min_val
    return amplitude


def _get_area(waveform, idx1, idx2):
    reflex_waveform = np.abs(waveform[idx1:idx2] - np.mean(waveform[idx1:idx2]))
    return np.trapz(reflex_waveform)


def _get_onset(waveform, reflex_idx, sd_win_idx):
    """Calculate onset of relex

    We use a threshold based on the amplitude of the background EMG activity
    and the s2pr.SD_MULTIPLIER (currently 3). However, the onset this is
    earlier than the point that crosses the threshold. It is the point where,
    going backwards from the crossing, the rectified EMG inflects back up.

    """
    abs_reflex = abs(waveform)
    abs_background_sd = np.std(abs_reflex[sd_win_idx[0]: sd_win_idx[1]])
    threshold = abs_background_sd * s2pr.utils.SD_MULTIPLIER

    # Determine if any value was greater than threshold
    threshold_crossed = sum(abs_reflex[reflex_idx[0]: reflex_idx[1]] > threshold)

    if threshold_crossed:
        crossing_idx = _get_crossing_idx(abs_reflex, reflex_idx, threshold)
        onset = _get_onset_idx(abs_reflex, crossing_idx)
        return onset, abs_background_sd
    else:
        return None, None


def _get_crossing_idx(abs_reflex, reflex_idx, threshold):
    for i in range(reflex_idx[0], reflex_idx[1]):
        if abs_reflex[i] > threshold:
            return i


def _get_onset_idx(abs_reflex, crossing_idx):
    for i in range(crossing_idx):
        if _inflexion_back_up(abs_reflex, crossing_idx):
            return crossing_idx - 1
        else:
            crossing_idx -= 1


def _inflexion_back_up(abs_reflex, crossing_idx):
    return abs_reflex[crossing_idx - 1] > abs_reflex[crossing_idx]



import matplotlib.pyplot as plt

import spike2py_reflex as s2pr

PLOT_WIDTH_SCALING = 1
PLOT_HEIGHT_SCALING = 2


def plot_double(section):
    """Plot section double reflexes.

    Generates a plot with n columns, where n = 2 * number_muscles, and m rows,
    where m = number of stimulation intensities. For each muscle, there are
    two columns. The first column plots the data extracted for the pair of reflexes,
    and also the mean waveform. The second column plots the same extracted reflex_1
    and reflex_2, overlayed in red and blue, as well as their averaged. Also provided
    on this plot data are the windows provided from which to compute reflex outcomes.

    """

    info = section.info
    muscles, n_muscles, intensities, n_intensities = s2pr.plot.get_muscles_and_intensities(section)
    plot_width, plot_height = s2pr.plot.get_plot_width_height(n_muscles, n_intensities)
    plt.figure(figsize=(plot_width * 2 * PLOT_WIDTH_SCALING, plot_height * 2))
    _gen_figure(muscles, n_muscles, intensities, n_intensities, section, info)
    plt.tight_layout()
    s2pr.plot.save_figure(section)
    plt.close()


def _gen_figure(muscles, n_muscles, intensities, n_intensities, section, info):
    """For each intensity of each muscle, add subplots"""
    subplot = 1
    for intensity in intensities:
        for muscle in muscles:
            # Extracted reflex
            plt.subplot(n_intensities, n_muscles * 2, subplot)
            s2pr.plot.overlay_reflexes(section, muscle, intensity)

            # reflex1 and reflex2
            plt.subplot(n_intensities, n_muscles * 2, subplot + 1)
            _overlay_double_reflexes(section, muscle, info, intensity)
            s2pr.plot.add_reflex_windows(section.info, muscle)
            s2pr.plot.set_x_lim(section.info.windows.ms_plotting.double_single_pulse)
            subplot += 2  # To skip to the plot of reflex1 and reflex2


def _overlay_double_reflexes(section, muscle, info, intensity=None):

    # Plot individual raw reflexes
    for reflex in section.reflexes[muscle].reflexes:
        if intensity is None:

            try:
                plt.plot(section.reflexes[muscle].x_axis_singles,
                         reflex.reflex1.waveform,
                         alpha=0.3,
                         color='red')
            except AttributeError:
                print('Plotting double reflexes: no reflex1')

            try:
                plt.plot(section.reflexes[muscle].x_axis_singles,
                         reflex.reflex2.waveform,
                         alpha=0.3,
                         color='blue')
            except AttributeError:
                print('Plotting double reflexes: no reflex2')
        else:
            if intensity == reflex.stim_intensity:
                try:
                    plt.plot(section.reflexes[muscle].x_axis_singles,
                             reflex.reflex1.waveform,
                             alpha=0.3,
                             color='red')
                except AttributeError:
                    print('Plotting double reflexes: no reflex1')

                try:
                    plt.plot(section.reflexes[muscle].x_axis_singles,
                             reflex.reflex2.waveform,
                             alpha=0.3,
                             color='blue')
                except AttributeError:
                    print('Plotting double reflexes: no reflex2')

    # Plot avg reflexes
    if intensity is None:
        try:
            for _, value in section.reflexes[muscle].avg_reflex1.items():
                plt.plot(section.reflexes[muscle].x_axis_singles,
                         value.waveform,
                         alpha=1,
                         color='red',
                         label='reflex1')
        except AttributeError:
            print('Plotting double reflexes: no avg reflex1')

        try:
            for _, value in section.reflexes[muscle].avg_reflex2.items():
                plt.plot(section.reflexes[muscle].x_axis_singles,
                         value.waveform,
                         alpha=1,
                         color='blue',
                         label='reflex2')
        except AttributeError:
            print('Plotting double reflexes: no avg reflex2')

    else:
        if section.reflexes[muscle].avg_reflex1 is not None:
            try:
                try:
                    avg_reflex = section.reflexes[muscle].avg_reflex1[intensity]
                    plt.plot(section.reflexes[muscle].x_axis_singles,
                             avg_reflex.waveform,
                             alpha=1,
                             color='red',
                             label='reflex1')
                except KeyError:
                    pass
            except AttributeError:
                print('Plotting double reflexes: no avg reflex1')

            try:
                try:
                    avg_reflex = section.reflexes[muscle].avg_reflex2[intensity]
                    plt.plot(section.reflexes[muscle].x_axis_singles,
                             avg_reflex.waveform,
                             alpha=1,
                             color='blue',
                             label='reflex2')
                except KeyError:
                    pass
            except AttributeError:
                print('Plotting double reflexes: no avg reflex2')

    plt.xlabel('time (s)')
    plt.grid()
    if intensity is None:
        plt.title((f'{info.subject}-'
                   f'{info.trial}-'
                   f'{info.section}-'
                   f'{muscle}-'
                   f'reflex1/2'), fontsize=10)
    else:
        plt.title((f'{info.subject}-'
                   f'{info.trial}-'
                   f'{info.section}-'
                   f'{muscle}-'
                   f'{intensity}mA-'
                   f'reflex1-2'), fontsize=10)



import matplotlib.pyplot as plt
from dataclasses import dataclass

import spike2py_reflex as s2pr


def plot_outcomes(section):
    """Generate raw data plots of all reflexes in a trial section

    Included are the individual traces and the mean trace, as well
    as windows indicating the region where reflex outcomes were
    calculated.
    """

    muscles = list(section.reflexes.keys())
    if section.reflexes[muscles[0]].type == s2pr.utils.SINGLE:
#        s2pr.plot.single_outcomes(section)
        single_outcomes(section)
    elif section.reflexes[muscles[0]].type == s2pr.utils.TRAIN:
#        s2pr.plot.train_outcomes(section)
        train_outcomes(section)
    elif section.reflexes[muscles[0]].type == s2pr.utils.DOUBLE:
#        s2pr.plot.double_outcomes(section)
       double_outcomes(section)


N_OUTCOMES = 3
WIDTH_SCALING = 12
HEIGHT_SCALING = 4
PEAK_TO_PEAK_COLOR = 'tab:pink'
PEAK_TO_PEAK_LABEL = 'peak-to-peak (a.u.)'
AREA_COLOR = 'tab:cyan'
AREA_LABEL = 'area (a.u.)'
ONSET_COLOR = 'tab:gray'
ONSET_LABEL = 'latency (ms)'
X_AXIS_LABEL = 'stim intensities'
P2P = 'peak_to_peak'
ONSET = 'onset'
AREA = 'area'
RATIO_LABEL = 'ratio (reflex2/reflex1)'
RATIO = 'ratio'
RATIO_COLOR = 'tab:orange'
SUBPLOT = {'peak_to_peak': 1,
           'area': 2,
           'onset': 3,
           'ratio': 4}

@dataclass
class SubplotInfo:
    rows: int
    cols: int
    row_counter: int
    muscle_subplot: int
    x: int
    y: float
    y_axis_label: str
    muscle: str
    reflex_type: str
    x_ticks: list
    color: str
    outcome: str
    num_outcomes: int
    x_axis_label: str = ''
    fillstyle: str = 'none'
    alpha: float = 1.0

def single_outcomes(section):
    muscles, n_muscles, intensities, n_intensities = s2pr.plot.get_muscles_and_intensities(
        section)
    n_reflex_type = len(section.info.windows.ms_reflexes[muscles[0]][section.info.section].keys())

    plt.figure(figsize=(WIDTH_SCALING*n_muscles, HEIGHT_SCALING*n_reflex_type))

    rows = n_reflex_type
    cols = n_muscles*N_OUTCOMES

    for muscle_subplot, muscle in enumerate(muscles):
        for intensity in intensities:
            for reflex in section.reflexes[muscle].reflexes:
                if intensity is None:
                    for row_counter, (reflex_type, outcomes) in enumerate(reflex.outcomes.items()):
                        # peak-to-peak
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=0,
                                                   y=outcomes.peak_to_peak,
                                                   y_axis_label=PEAK_TO_PEAK_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=list(),
                                                   outcome=P2P,
                                                   num_outcomes=3,
                                                   color=PEAK_TO_PEAK_COLOR)
                        subplot(subplot_info)

                        # area
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=0,
                                                   y=outcomes.area,
                                                   y_axis_label=AREA_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=list(),
                                                   outcome=AREA,
                                                   num_outcomes=3,
                                                   color=AREA_COLOR)
                        subplot(subplot_info)

                        # onset
                        if outcomes.onset is not None:
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=0,
                                                       y=outcomes.area,
                                                       y_axis_label=ONSET_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=list(),
                                                       outcome=ONSET,
                                                       num_outcomes=3,
                                                       color=ONSET_COLOR)
                            subplot(subplot_info)


                if intensity == reflex.stim_intensity:
                    for row_counter, (reflex_type, outcomes) in enumerate(
                            reflex.outcomes.items()):
                        # peak-to-peak
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=intensity,
                                                   y=outcomes.peak_to_peak,
                                                   y_axis_label=PEAK_TO_PEAK_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=intensities,
                                                   x_axis_label=X_AXIS_LABEL,
                                                   outcome=P2P,
                                                   num_outcomes=3,
                                                   color=PEAK_TO_PEAK_COLOR)
                        subplot(subplot_info)

                        # area
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=intensity,
                                                   y=outcomes.area,
                                                   y_axis_label=AREA_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=intensities,
                                                   x_axis_label=X_AXIS_LABEL,
                                                   outcome=AREA,
                                                   num_outcomes=3,
                                                   color=AREA_COLOR)
                        subplot(subplot_info)

                        # onset
                        if outcomes.onset is not None:
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=intensity,
                                                       y=outcomes.onset,
                                                       y_axis_label=ONSET_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=intensities,
                                                       x_axis_label=X_AXIS_LABEL,
                                                       outcome=ONSET,
                                                       num_outcomes=3,
                                                       color=ONSET_COLOR)
                            subplot(subplot_info)
    plt.tight_layout()
    save_figure(section)
    plt.close()


def subplot(sp):
    subplot_id = (sp.row_counter * sp.cols) + (sp.muscle_subplot * sp.num_outcomes) + SUBPLOT[sp.outcome]
    plt.subplot(sp.rows, sp.cols, subplot_id)
    plt.plot(sp.x, sp.y, 'o', fillstyle=sp.fillstyle, color=sp.color,
             alpha=sp.alpha)
    plt.ylabel(sp.y_axis_label)
    plt.title(f'{sp.muscle}-{sp.reflex_type}')
    if len(sp.x_ticks) < 8:
        plt.xticks(ticks=sp.x_ticks)
    elif len(sp.x_ticks) < 15:
        plt.xticks(ticks=sp.x_ticks, fontsize=9)
    else:
        plt.xticks(ticks=sp.x_ticks, fontsize=6)
    plt.xlabel(sp.x_axis_label)



def save_figure(section):
    plt.suptitle(f'{section.info.trial}_{section.info.section}_outcomes')
    plt.tight_layout()
    figure_name = f'{section.info.trial}_{section.info.section}_outcomes.pdf'
    path = section.info.study_path / section.info.subject / 'figures' / 'outcomes'
    if not path.exists():
        path.mkdir()
    plt.savefig(path/figure_name)


def train_outcomes(section):
    muscles, n_muscles, intensities, n_intensities = s2pr.plot.get_muscles_and_intensities(
        section)
    n_reflex_type = len(section.info.windows.ms_reflexes[muscles[0]][section.info.section].keys())

    plt.figure(figsize=(WIDTH_SCALING*0.75*n_muscles, HEIGHT_SCALING*n_reflex_type))

    rows = n_reflex_type
    cols = n_muscles*(N_OUTCOMES-1)

    for muscle_subplot, muscle in enumerate(muscles):
        for intensity in intensities:
            for reflex in section.reflexes[muscle].reflexes:
                if intensity is None:
                    for row_counter, (reflex_type, outcomes) in enumerate(reflex.outcomes.items()):
                        # peak-to-peak
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=0,
                                                   y=outcomes.peak_to_peak,
                                                   y_axis_label=PEAK_TO_PEAK_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=list(),
                                                   outcome=P2P,
                                                   num_outcomes=2,
                                                   color=PEAK_TO_PEAK_COLOR)
                        subplot(subplot_info)

                        # area
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=0,
                                                   y=outcomes.area,
                                                   y_axis_label=AREA_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=list(),
                                                   outcome=AREA,
                                                   num_outcomes=2,
                                                   color=AREA_COLOR)
                        subplot(subplot_info)

                if intensity == reflex.stim_intensity:
                    for row_counter, (reflex_type, outcomes) in enumerate(
                            reflex.outcomes.items()):
                        # peak-to-peak
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=intensity,
                                                   y=outcomes.peak_to_peak,
                                                   y_axis_label=PEAK_TO_PEAK_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=intensities,
                                                   x_axis_label=X_AXIS_LABEL,
                                                   outcome=P2P,
                                                   num_outcomes=2,
                                                   color=PEAK_TO_PEAK_COLOR)
                        subplot(subplot_info)

                        # area
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=intensity,
                                                   y=outcomes.area,
                                                   y_axis_label=AREA_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=intensities,
                                                   x_axis_label=X_AXIS_LABEL,
                                                   outcome=AREA,
                                                   num_outcomes=2,
                                                   color=AREA_COLOR)
                        subplot(subplot_info)
    plt.tight_layout()
    save_figure(section)
    plt.close()


def double_outcomes(section):
    muscles, n_muscles, intensities, n_intensities = s2pr.plot.get_muscles_and_intensities(
        section)
    n_reflex_type = len(section.info.windows.ms_reflexes[muscles[0]][section.info.section].keys())

    plt.figure(figsize=(WIDTH_SCALING*n_muscles*1.25, HEIGHT_SCALING*n_reflex_type))

    rows = n_reflex_type
    cols = n_muscles*(N_OUTCOMES+1)

    for muscle_subplot, muscle in enumerate(muscles):
        for intensity in intensities:
            for reflex in section.reflexes[muscle].reflexes:
                if intensity is None:
                    # Reflex 1
                    for row_counter, (reflex_type, outcomes) in enumerate(reflex.reflex1.outcomes.items()):
                        # peak-to-peak
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=0,
                                                   y=outcomes.peak_to_peak,
                                                   y_axis_label=PEAK_TO_PEAK_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=list(),
                                                   outcome=P2P,
                                                   num_outcomes=4,
                                                   color=PEAK_TO_PEAK_COLOR)
                        subplot(subplot_info)

                        # area
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=0,
                                                   y=outcomes.area,
                                                   y_axis_label=AREA_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=list(),
                                                   outcome=AREA,
                                                   num_outcomes=4,
                                                   color=AREA_COLOR)
                        subplot(subplot_info)

                        # onset
                        if outcomes.onset is not None:
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=0,
                                                       y=outcomes.onset,
                                                       y_axis_label=ONSET_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=list(),
                                                       outcome=ONSET,
                                                       num_outcomes=3,
                                                       color=ONSET_COLOR)
                            subplot(subplot_info)

                    # Reflex 2
                    try:  # In case there is not reflex 2
                        for row_counter, (reflex_type, outcomes) in enumerate(
                                reflex.reflex2.outcomes.items()):
                            # peak-to-peak
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=1,
                                                       y=outcomes.peak_to_peak,
                                                       y_axis_label=PEAK_TO_PEAK_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=list(),
                                                       outcome=P2P,
                                                       num_outcomes=4,
                                                       color=PEAK_TO_PEAK_COLOR,
                                                       fillstyle='full',
                                                       alpha=0.3)
                            subplot(subplot_info)

                            # area
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=1,
                                                       y=outcomes.area,
                                                       y_axis_label=AREA_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=list(),
                                                       outcome=AREA,
                                                       num_outcomes=4,
                                                       color=AREA_COLOR,
                                                       fillstyle='full',
                                                       alpha=0.3
                                                       )
                            subplot(subplot_info)

                            # onset
                            if outcomes.onset is not None:
                                subplot_info = SubplotInfo(rows=rows,
                                                           cols=cols,
                                                           row_counter=row_counter,
                                                           muscle_subplot=muscle_subplot,
                                                           x=1,
                                                           y=outcomes.onset,
                                                           y_axis_label=ONSET_LABEL,
                                                           muscle=muscle,
                                                           reflex_type=reflex_type,
                                                           x_ticks=list(),
                                                           outcome=ONSET,
                                                           num_outcomes=4,
                                                           color=ONSET_COLOR,
                                                           fillstyle='full',
                                                           alpha=0.3
                                                           )
                                subplot(subplot_info)
                    except AttributeError:
                        print('Plotting double outcomes: no reflex2')
                    # Ratio
                    try:
                        for row_counter, (reflex_type, ratio) in enumerate(
                                reflex.ratio.items()):
                            # ratio
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=0,
                                                       y=ratio,
                                                       y_axis_label=RATIO_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=list(),
                                                       outcome=RATIO,
                                                       num_outcomes=4,
                                                       color=RATIO_COLOR)
                            subplot(subplot_info)
                    except AttributeError:
                        print('Plotting double outcomes: no ratio outcome')

                if intensity == reflex.stim_intensity:
                    # Reflex 1
                    for row_counter, (reflex_type, outcomes) in enumerate(
                            reflex.reflex1.outcomes.items()):
                        # peak-to-peak
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=intensity,
                                                   y=outcomes.peak_to_peak,
                                                   y_axis_label=PEAK_TO_PEAK_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=intensities,
                                                   outcome=P2P,
                                                   num_outcomes=4,
                                                   x_axis_label=X_AXIS_LABEL,
                                                   color=PEAK_TO_PEAK_COLOR)
                        subplot(subplot_info)

                        # area
                        subplot_info = SubplotInfo(rows=rows,
                                                   cols=cols,
                                                   row_counter=row_counter,
                                                   muscle_subplot=muscle_subplot,
                                                   x=intensity,
                                                   y=outcomes.area,
                                                   y_axis_label=AREA_LABEL,
                                                   muscle=muscle,
                                                   reflex_type=reflex_type,
                                                   x_ticks=intensities,
                                                   outcome=AREA,
                                                   num_outcomes=4,
                                                   x_axis_label=X_AXIS_LABEL,
                                                   color=AREA_COLOR)
                        subplot(subplot_info)

                        # onset
                        if outcomes.onset is not None:
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=intensity,
                                                       y=outcomes.onset,
                                                       y_axis_label=ONSET_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=intensities,
                                                       outcome=ONSET,
                                                       num_outcomes=4,
                                                       x_axis_label=X_AXIS_LABEL,
                                                       color=ONSET_COLOR)
                            subplot(subplot_info)
                    # Reflex 2
                    try:
                        for row_counter, (reflex_type, outcomes) in enumerate(
                                reflex.reflex2.outcomes.items()):
                            # peak-to-peak
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=intensity + 1,
                                                       y=outcomes.peak_to_peak,
                                                       y_axis_label=PEAK_TO_PEAK_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=intensities,
                                                       outcome=P2P,
                                                       num_outcomes=4,
                                                       x_axis_label=X_AXIS_LABEL,
                                                       color=PEAK_TO_PEAK_COLOR,
                                                       fillstyle='full',
                                                       alpha=0.3
                                                       )
                            subplot(subplot_info)

                            # area
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=intensity + 1,
                                                       y=outcomes.area,
                                                       y_axis_label=AREA_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=intensities,
                                                       outcome=AREA,
                                                       num_outcomes=4,
                                                       x_axis_label=X_AXIS_LABEL,
                                                       color=AREA_COLOR,
                                                       fillstyle='full',
                                                       alpha=0.3
                                                       )
                            subplot(subplot_info)

                            # onset
                            if outcomes.onset is not None:
                                subplot_info = SubplotInfo(rows=rows,
                                                           cols=cols,
                                                           row_counter=row_counter,
                                                           muscle_subplot=muscle_subplot,
                                                           x=intensity + 1,
                                                           y=outcomes.onset,
                                                           y_axis_label=ONSET_LABEL,
                                                           muscle=muscle,
                                                           reflex_type=reflex_type,
                                                           x_ticks=intensities,
                                                           outcome=ONSET,
                                                           num_outcomes=4,
                                                           x_axis_label=X_AXIS_LABEL,
                                                           color=ONSET_COLOR,
                                                           fillstyle='full',
                                                           alpha=0.3
                                                           )
                                subplot(subplot_info)
                    except AttributeError:
                        print('Plotting double outcomes: no reflex2')

                    try:
                        # Ratio
                        for row_counter, (reflex_type, ratio) in enumerate(
                                reflex.ratio.items()):
                            # ratio
                            subplot_info = SubplotInfo(rows=rows,
                                                       cols=cols,
                                                       row_counter=row_counter,
                                                       muscle_subplot=muscle_subplot,
                                                       x=intensity,
                                                       y=ratio,
                                                       y_axis_label=RATIO_LABEL,
                                                       muscle=muscle,
                                                       reflex_type=reflex_type,
                                                       x_ticks=intensities,
                                                       outcome=RATIO,
                                                       num_outcomes=4,
                                                       x_axis_label=X_AXIS_LABEL,
                                                       color=RATIO_COLOR)
                            subplot(subplot_info)
                    except AttributeError:
                        print('Plotting double outcomes: no ratio outcome')
    plt.tight_layout()
    save_figure(section)
    plt.close()



import numpy as np
import spike2py_reflex as s2pr
import spike2py_preprocess as s2pp


def calculate_outcomes_of_avg(section):
    """Calculate avg reflex and its outcomes

       For each muscle and for each type of extracted reflex (e.g. H-reflex, mMax),
       calculate the mean waveform and the outcomes of this mean waveform.
       If stimulation was given at different intensities, each will have their
       own average waveform and outcomes.

       Returns
       -------
       For single:

        reflexes.avg_waveform[stim_intensity] = s2pr.Single(waveform=avg_reflex_waveform,
                                                            outcomes=all_outcomes,
                                                            background_sd=background_sd)

       For double:

        reflexes.avg_reflex1[stim_intensity] = s2pr.Single(waveform=avg_reflex_waveform[lower_idx1: upper_idx1],
                                                           outcomes=all_outcomes_reflex1,
                                                           background_sd=background_sd)

        reflexes.avg_reflex2[stim_intensity] = s2pr.Single(waveform=avg_reflex_waveform[lower_idx2: upper_idx2],
                                                           outcomes=all_outcomes_reflex2,
                                                           background_sd=background_sd)
       """

    for muscle, reflexes in section.reflexes.items():
        stim_intensities = get_stim_intensities(reflexes)

        if reflexes.type in [s2pr.utils.SINGLE, s2pr.utils.TRAIN]:
            section.reflexes[muscle] = _single_calculate_avg(reflexes, stim_intensities, section.info.section)

        elif reflexes.type == s2pr.utils.DOUBLE:
            info = section.info
            section.reflexes[muscle] = _double_calculate_avg(reflexes, stim_intensities, info)

    return section


def _single_calculate_avg(reflexes, stim_intensities, section_name):
    reflexes.avg_waveform = dict()
    if _no_stim_intensities(stim_intensities):
        reflexes = _get_single_avg(reflexes, stim_intensities, section_name)
    else:
        for stim_intensity in stim_intensities:
            reflexes = _get_single_avg(reflexes, stim_intensity, section_name)
    return reflexes


def _no_stim_intensities(stim_intensities):
    return not stim_intensities


def get_stim_intensities(reflexes):
    stim_intensities = list()
    for reflex in reflexes.reflexes:
        if reflex.stim_intensity:
            stim_intensities.append(reflex.stim_intensity)
    if len(stim_intensities) == 0:
        return None
    return set(stim_intensities)


def _get_single_avg(reflexes, stim_intensity, section_name):

    # Get average waveform
    avg_reflex_waveform = list()
    for reflex in reflexes.reflexes:
        if (reflex.stim_intensity == stim_intensity) or (stim_intensity is None):
            avg_reflex_waveform.append(reflex.waveform)
    avg_reflex_waveform = np.array(avg_reflex_waveform).mean(axis=0)

    # Get values needed to compute outcomes
    x_axis = reflexes.x_axis_extract

    sd_idx_all_stim_times = reflexes.sd_window_idx
    # Get idx windows of all reflexes types (e.g. H-reflex + mMax; or just mMax)
    reflex_win_idx_all = reflexes.reflex_windows_idx[section_name]

    # Compute outcomes for each reflex type
    all_outcomes = dict()
    for reflex_type, reflex_idx_dict in reflex_win_idx_all.items():
        reflex_win_idx = reflex_idx_dict[reflexes.type]
        if reflexes.type == s2pr.utils.SINGLE:
            sd_idx = sd_idx_all_stim_times[reflexes.type]
        else:
            sd_idx = None
        outcomes, background_sd = s2pr.outcomes.get_outcomes_from(avg_reflex_waveform,
                                                         reflex_win_idx,
                                                         sd_idx, x_axis)
        all_outcomes[reflex_type] = outcomes

    if stim_intensity is None:
        stim_intensity = 'no_intensity'
    reflexes.avg_waveform[stim_intensity] = s2pr.reflexes.Single(waveform=avg_reflex_waveform,
                                                        outcomes=all_outcomes,
                                                        background_sd=background_sd)
    return reflexes


def _double_calculate_avg(reflexes, stim_intensities, info):
    reflexes.avg_waveform = dict()
    reflexes.avg_waveform_for_doubles = dict()

    reflexes.avg_reflex1 = dict()
    reflexes.avg_reflex1_for_doubles = dict()

    reflexes.avg_reflex2 = dict()

    if _no_stim_intensities(stim_intensities):
        reflexes = _get_double_avg(reflexes, stim_intensities, info)
    else:
        for stim_intensity in stim_intensities:
            reflexes = _get_double_avg(reflexes, stim_intensity, info)
    return reflexes


def _get_double_avg(reflexes, stim_intensity, info):

    # Get average waveform
    avg_reflex_waveform = list()  # avg of all waveforms (regardless if mix of Single and Double)
    avg_reflex_waveform_for_doubles = list()  # avg of all waveforms, but excluding those that were a Single
    for reflex in reflexes.reflexes:
        if (reflex.stim_intensity == stim_intensity) or (stim_intensity is None):
            avg_reflex_waveform.append(reflex.waveform)
            if reflex.reflex2 is not None:
                # Do not include instances of Singles when computing avg waveform for Doubles
                avg_reflex_waveform_for_doubles.append(reflex.waveform)

    avg_reflex_waveform = np.array(avg_reflex_waveform).mean(axis=0)
    avg_reflex_waveform_for_doubles = np.array(avg_reflex_waveform_for_doubles).mean(axis=0)

    # avg_waveform and avg_waveform_for_doubles are the same if all reflexes were Double
    # If there was a mix of Single and Double:
    #    avg_waveform_for_doubles will be the avg of the waveforms where there was both a reflex 1 and reflex 2
    #    avg_waveform will the avg of all the waveforms; here, the window where reflex 2 is incorrect as it
    #    represents the avg of one or more true reflex 2 and background noise (in the case of Single).
    reflexes.avg_waveform[stim_intensity] = avg_reflex_waveform
    reflexes.avg_waveform_for_doubles[stim_intensity] = avg_reflex_waveform_for_doubles

    # Get values needed to compute outcomes
    x_axis = reflexes.x_axis_extract

    sd_idx_all_stim_times = reflexes.sd_window_idx
    # Get idx windows of all reflexes types (e.g. H-reflex + mMax; or just mMax)
    reflex_win_idx_all = reflexes.reflex_windows_idx[info.section]

    # Compute outcomes for reflex(es)
    all_outcomes_reflex1 = dict()
    all_outcomes_reflex1_for_doubles = dict()
    all_outcomes_reflex2 = dict()

    for reflex_type, reflex_idx_dict in reflex_win_idx_all.items():
        reflex_win_idx = reflex_idx_dict[reflexes.type]
        sd_idx = sd_idx_all_stim_times[reflexes.type]
        # compute outcomes for reflex1 when avg waveform includes all reflexes (i.e. Single and Double)
        outcomes_reflex1, background_sd = s2pr.outcomes.get_outcomes_from(avg_reflex_waveform, reflex_win_idx[0], sd_idx, x_axis)
        all_outcomes_reflex1[reflex_type] = outcomes_reflex1

        # compute outcomes for reflex1 when avg waveform only includes all the Double
        outcomes_reflex1_for_doubles, background_sd_for_doubles = s2pr.outcomes.get_outcomes_from(avg_reflex_waveform_for_doubles, reflex_win_idx[0], sd_idx, x_axis)
        all_outcomes_reflex1_for_doubles[reflex_type] = outcomes_reflex1_for_doubles

        if reflex_win_idx[1] is not None:
            outcomes_reflex2, _ = s2pr.outcomes.get_outcomes_from(avg_reflex_waveform_for_doubles, reflex_win_idx[1], sd_idx, x_axis)
            all_outcomes_reflex2[reflex_type] = outcomes_reflex2

    if stim_intensity is None:
        stim_intensity = 'no_intensity'

    idx_double_single_pulse = info.windows.idx_extract.double_single_pulse
    idx_time_zero = s2pp.find_nearest_time_index(x_axis, 0)

    lower_idx1 = idx_time_zero + idx_double_single_pulse[0]
    upper_idx1 = idx_time_zero + idx_double_single_pulse[1]

    idx_time_isi = s2pp.find_nearest_time_index(x_axis, info.windows.double_isi * s2pr.utils.CONVERT_MS_TO_S)

    lower_idx2 = idx_time_isi + idx_double_single_pulse[0]
    upper_idx2 = idx_time_isi + idx_double_single_pulse[1]

    reflexes.avg_reflex1[stim_intensity] = s2pr.reflexes.Single(waveform=avg_reflex_waveform[lower_idx1: upper_idx1],
                                                        outcomes=all_outcomes_reflex1,
                                                        background_sd=background_sd)

    reflexes.avg_reflex1_for_doubles[stim_intensity] = s2pr.reflexes.Single(waveform=avg_reflex_waveform_for_doubles[lower_idx1: upper_idx1],
                                                                outcomes=all_outcomes_reflex1_for_doubles,
                                                                background_sd=background_sd_for_doubles)

    try:
        reflexes.avg_reflex2[stim_intensity] = s2pr.reflexes.Single(waveform=avg_reflex_waveform_for_doubles[lower_idx2: upper_idx2],
                                                           outcomes=all_outcomes_reflex2,
                                                           background_sd=background_sd_for_doubles)
    except AttributeError:
        pass
    return reflexes



import numpy as np
from spike2py_preprocess.trial_sections import find_nearest_time_index

import spike2py_reflex as s2pr


class Triggers:

    def __init__(self, info, data):
        """

        Parameters
        ----------
        info
        data

        Attributes
        ----------
        extract: list of indexes corresponding to each trigger; for double stim it corresponds to the first trigger
        double: list of pairs of indexes corresponding to the pair of triggers in a double
        """

        section = info.section
        self.type = info.channels.triggers[section]["type"]

        trigger_channel = getattr(data, info.channels.triggers[section]["channels"])
        self._raw = trigger_channel.times

        emg = getattr(data, info.channels.emg[0])
        self._times = emg.times

        self._triggers = self._clean_triggers(info)

        self.extract = None
        self.double = None

        self._get_idx(info)

    def _clean_triggers(self, info):
        triggers = self._remove_khz()
        return self._remove_rejected_triggers(triggers, info)

    def _remove_khz(self):
        triggers = [self._raw[0]]
        for i in range(1, len(self._raw)):
            trig_isi = self._raw[i] - self._raw[i - 1]
            if trig_isi > s2pr.utils.KHZ_ISI_THRESHOLD:
                triggers.append(self._raw[i])
        return triggers

    def _remove_rejected_triggers(self, triggers: list, info) -> list:
        """Remove triggers flagged for removal.

        `rejected_triggers_times` are extracted from the section-specific JSON file.
        If no triggers flagged for removal, `triggers` is returned unmodified.

        Parameters
        ----------
        triggers: Trigger times.
        rejected_trigger_times: List of value pairs indicating windows in which to remove triggers.
                                Example: [[5.5, 12], [45, 45.3]]
                                Triggers present between 5.5s and 12s, and those present between
                                45s and 45.3s would be removed.

        Returns
        -------
        List of trigger times.
        """
        if info.rejected_trigger_windows:
            original_triggers = np.array(triggers)

            trigger_mask = np.bool_(np.ones(len(triggers)))  # Initial mask `True` for all triggers
            for lower_value, upper_value in info.rejected_trigger_windows:
                lower = original_triggers > lower_value
                upper = original_triggers < upper_value
                trigger_mask *= np.invert(lower * upper)  # Update mask based on current trigger times
            triggers = original_triggers[trigger_mask]    # Use mask to retain only wanted triggers.
        return triggers

    def _get_idx(self, info):
        if self.type in [s2pr.utils.SINGLE, s2pr.utils.TRAIN]:
            self.extract = self._get_idx_single_triggers()
        elif self.type == s2pr.utils.DOUBLE:
            extract, double = self._get_idx_double_triggers(info)
            self.extract = extract
            self.double = double

    def _get_idx_single_triggers(self):
        """Get idx of each trigger."""
        trigger_indexes = list()
        for trigger in self._triggers:
            trigger_indexes.append(find_nearest_time_index(self._times, trigger))
        return trigger_indexes

    def _get_idx_double_triggers(self, info):
        """Get indexes to extract data for entire double and each individual reflex.

        It is possible that doubles were used at the start of section, followed by singles.
        Need to account for this possibility. If it happens, the single reflexes will be
        stored as reflex_1 of dc.ReflexDouble.reflex_1.
        """
        extract = list()
        double = list()
        # If trig1-trig2 have double isi, skip one pulse - trig2 cannot be start of new double.
        skip_one = False
        for i in range(len(self._triggers)):
            if not skip_one:
                trig1 = self._triggers[i]
                try:
                    # Last trigger can be single or part of double.
                    # If part of double, can access triggers[i + 1], set skip_one to True and then end.
                    # If not, last trigger will be trig1 (of single) and triggers[i + 1] will give
                    # AssertionError.
                    trig2 = self._triggers[i + 1]
                    isi = trig2 - trig1
                    try:
                        # If isi not approximately double_isi, AssertionError
                        self._is_double(isi, info)
                        extract.append(find_nearest_time_index(self._times, trig1))
                        double1 = find_nearest_time_index(self._times, trig1)
                        double2 = find_nearest_time_index(self._times, trig2)
                        double.append([double1, double2])
                        skip_one = True
                    except AssertionError:
                        # ISI not compatible with double, thus must be a single.
                        extract.append(find_nearest_time_index(self._times, trig1))
                        double1 = find_nearest_time_index(self._times, trig1)
                        double.append([double1, None])
                except IndexError:
                    # Tried to access trigger after last available trigger.
                    # Means last trigger is for a single.
                    extract.append(find_nearest_time_index(self._times, trig1))
                    double1 = find_nearest_time_index(self._times, trig1)
                    double.append([double1, None])
            else:
                skip_one = False
        return extract, double

    def _is_double(self, isi, info):
        double_isi = info.stim_params.double_isi * s2pr.utils.CONVERT_MS_TO_S
        return np.testing.assert_approx_equal(actual=isi, desired=double_isi, significant=3)
